from ctypes import *
from unittest import TestCase
from .wraped_kernel32 import Kernel32, FailedKernel32, MAX_COMPUTERNAME_LENGTH
from wrapdll import DllError
from socket import gethostname

class WrapedKernel32TestCase(TestCase):

    def setUp(self):
        self.__dll = Kernel32()

    def tearDown(self):
        del self.__dll

    def test_GetCurrentProcess(self):
        self.assertEqual(self.__dll.GetCurrentProcess(), wintypes.HANDLE(-1).value)

    def test_IsWow64Process(self):
        current_process = self.__dll.GetCurrentProcess()
        import platform
        python_arch = platform.architecture()[0]
        PFALSE = pointer(wintypes.BOOL(False))
        is_wow = self.__dll.IsWow64Process(current_process, PFALSE)
        self.assertEqual(is_wow, python_arch=='32bit')

    def test_GetComputerNameA(self):
        lpBuffer = create_string_buffer(MAX_COMPUTERNAME_LENGTH)
        size = len(lpBuffer)
        p_size = pointer(wintypes.DWORD(size))
        is_success = self.__dll.GetComputerNameA(lpBuffer, p_size)
        real_hostname = gethostname()
        self.assertTrue(is_success)
        self.assertEqual(p_size.contents.value, len(real_hostname))
        self.assertEqual(lpBuffer.value.decode('utf8'), real_hostname)

class WrapedFailedKernel32TestCase(TestCase):

    def setUp(self):
        self.__dll = FailedKernel32()

    def tearDown(self):
        del self.__dll

    def test_forbidden_values(self):
        with self.assertRaises(DllError):
            self.__dll.GetCurrentProcess()



from ctypes import *
from unittest import TestCase
from sys import getrefcount as grc
from .wraped_python_api import PythonAPI

class WrapedPythonAPITestCase(TestCase):
    """
    Copied from ctypes test_python_api.py
    """
    dll = None

    @classmethod
    def setUpClass(cls):
        cls.dll = PythonAPI()

    def test_PyBytes_FromStringAndSize(self):
        PyBytes_FromStringAndSize = self.dll.PyBytes_FromStringAndSize

        # PyBytes_FromStringAndSize.restype = py_object
        # PyBytes_FromStringAndSize.argtypes = c_char_p, c_py_ssize_t

        self.assertEqual(PyBytes_FromStringAndSize(b"abcdefghi", 3), b"abc")

    def test_PyString_FromString(self):
        # pythonapi.PyBytes_FromString.restype = py_object
        # pythonapi.PyBytes_FromString.argtypes = (c_char_p,)

        s = b"abc"
        refcnt = grc(s)
        pyob = self.dll.PyBytes_FromString(s)
        self.assertEqual(grc(s), refcnt)
        self.assertEqual(s, pyob)
        del pyob
        self.assertEqual(grc(s), refcnt)

    def test_PyLong_Long(self):
        ref42 = grc(42)
        # pythonapi.PyLong_FromLong.restype = py_object
        self.assertEqual(self.dll.PyLong_FromLong(42), 42)

        self.assertEqual(grc(42), ref42)

        # pythonapi.PyLong_AsLong.argtypes = (py_object,)
        # pythonapi.PyLong_AsLong.restype = c_long

        res = self.dll.PyLong_AsLong(42)
        self.assertEqual(grc(res), ref42 + 1)
        del res
        self.assertEqual(grc(42), ref42)

    def test_PyOS_snprintf(self):
        PyOS_snprintf = self.dll.PyOS_snprintf
        # PyOS_snprintf.argtypes = POINTER(c_char), c_size_t, c_char_p

        buf = c_buffer(256)
        PyOS_snprintf(buf, sizeof(buf), b"Hello from %s", b"ctypes")
        self.assertEqual(buf.value, b"Hello from ctypes")

        PyOS_snprintf(buf, sizeof(buf), b"Hello from %s (%d, %d, %d)", b"ctypes", 1, 2, 3)
        self.assertEqual(buf.value, b"Hello from ctypes (1, 2, 3)")

        # not enough arguments
        self.assertRaises(TypeError, PyOS_snprintf, buf)



"""
Help wrap dll using python annotations as type hints
"""
from ctypes import cdll
from functools import wraps
from enum import IntEnum


def wrapdll(method):
    """
    Used to wrap methods of BaseDllWrapper's subcasses.
    Will replace the method with a dll exported function with the same name
    while using the annotations as typing hints 
    """
    @wraps(method)
    def _wrapdll_impl(self, *method_args, **method_kwargs):
        dll_func = getattr(self._dll, method.__name__)
        types = dict(method.__annotations__)
        # Add restype from annotations
        if 'return' in types:
            dll_func.restype = types['return']
            del types['return']
        # Add argtypes from annotations
        dll_func.argtypes = list(types.values())
        # Add defaults valeus if exists
        dll_func.__defaults__ = method.__defaults__
        dll_func.__kwdefaults__ = method.__kwdefaults__
        # Add errcheck func from the errcheck decorator
        errcheck = method_kwargs.pop('_errcheck_', None)
        if errcheck is not None:
            dll_func.errcheck = errcheck
        return dll_func(*method_args, **method_kwargs)
    return _wrapdll_impl

class BaseDllWrapper():
    """
    Derive from this class in conjucture with wrapdll decorator to wrap dlls
    """

    def __init__(self, ctypes_dll):
        """
        :param ctypes_dll: A ctypes dll (cdll or windll) or a str.
                           If got a str then tries to convert it to a cdll
        """
        if isinstance(ctypes_dll, str):
            ctypes_dll = cdll(ctypes_dll)
        self._dll = ctypes_dll

class RCEnum(IntEnum):
    """
    Derive from this cass to create return code enums with a msg.
    Ex:
    class CustomRC(RCEnum):
        SUCCESS = 0x00, 'No problem'
        CONNECTION_ERROR = 0x01, 'Failed to connect!'
        NO_MSG = 0x03
    """
    def __new__(cls, value, msg=None):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.msg = msg
        return obj

    def __repr__(self):
        return "<{}.{}: 0x{:X}>".format(self.__class__.__name__, self._name_, self._value_)

    def __str__(self):
        if self.msg:
            return "{}.{}: {}".format(self.__class__.__name__, self._name_, self.msg)
        return "{}.{}".format(self.__class__.__name__, self._name_)

def errcheck(err_func):
    """
    Adds an errcheck callable to a wrapdll method.
    err_func signature should be err_func(result, func, arguments) (Read ctypes doc for more)
    Notice the wrapdll decorator must be bellow this decorator.
    EX:
        def validate(result, func, arguments):
            if result:
                raise DllError(result)
        @errcheck(validate)
        @wrapdll
        def func():
            pass
    """
    def decorator(method):
        @wraps(method)
        def _errcheck_impl(self, *method_args, **method_kwargs):
            method_kwargs['_errcheck_'] = err_func
            return method(self, *method_args, **method_kwargs)
        return _errcheck_impl
    return decorator

class DllError(Exception):
    pass

def allowed_values(*values, exception=DllError):
    """
    Like errcheck decorator, but instead of a callable gets allowed values.
    Useful with RCEnum to check for invalid return codes
    """
    return _assert_values(*values, exception=exception, is_allowed=False)

def forbidden_values(*values, exception=DllError):
    """
    Like errcheck decorator, but instead of a callable gets forbidden values.
    Useful with RCEnum to check for invalid return codes
    """
    return _assert_values(*values, exception=exception, is_allowed=True)

def _assert_values(*values, exception=DllError, is_allowed=True):
    def _assert(result, func, arguments):
        # restype automatically dereference pointers, so add dereferenced pointers from values 
        dereferences = [value.value for value in values if hasattr(value, 'value')]
        dereferences.extend(values)
        if is_allowed:
            # Treat as allowed_values
            check = result in dereferences
        else:
            # Treat as forbidden_values
            check = result not in dereferences
        if check:
            raise exception(result)
        return result
    return errcheck(_assert)





from ctypes import *
from wrapdll import BaseDllWrapper, wrapdll, forbidden_values, allowed_values


MAX_COMPUTERNAME_LENGTH = 32

class Kernel32(BaseDllWrapper):
    """
    A wrapper to ctypes.windll.kernel32
    """

    def __init__(self):
        # Doesn't use windll.kernel32 in order to not change it
        super().__init__(windll.LoadLibrary('kernel32'))

    @wrapdll
    def GetCurrentProcess(self) -> wintypes.HANDLE:
        """
        Retrieves a pseudo handle for the current process.
        """
        pass

    @forbidden_values(0, exception=lambda x: WinError())
    @wrapdll
    def IsWow64Process(self, hProcess: wintypes.HANDLE, Wow64Process: wintypes.PBOOL) -> wintypes.BOOL:
        """
        Determines whether the specified process is running under WOW64 or an Intel64 of x64 processor.
        """
        pass

    @forbidden_values(0, exception=lambda x: WinError())
    @wrapdll
    def GetComputerNameA(lpBuffer: wintypes.LPSTR, nSize: wintypes.LPDWORD) -> wintypes.BOOL:
        """
        Retrieves the NetBIOS name of the local computer. 
        This name is established at system startup, when the system reads it from the registry.
        """
        pass

class FailedKernel32(BaseDllWrapper):
    def __init__(self):
        # Doesn't use windll.kernel32 in order to not change it
        super().__init__(windll.LoadLibrary('kernel32'))

    @forbidden_values(wintypes.HANDLE(-1))
    @wrapdll
    def GetCurrentProcess(self) -> wintypes.HANDLE:
        """
        Retrieves a pseudo handle for the current process.
        """
        pass


from ctypes import *
from wrapdll import BaseDllWrapper, wrapdll 

class PythonAPI(BaseDllWrapper):
    """
    A wrapper to ctypes.pythonapi
    """

    def __init__(self):
        super().__init__(pythonapi)

    @wrapdll
    def PyBytes_FromStringAndSize(self, v: c_char_p, len: c_size_t) -> py_object:
        """
        Return a new bytes object with a copy of the string v as value and length len on success, and NULL on failure.
        If v is NULL, the contents of the bytes object are uninitialized.
        """
        pass

    @wrapdll
    def PyBytes_FromString(self, v: c_char_p) -> py_object:
        """
        Return a new bytes object with a copy of the string v as value on success, and NULL on failure.
        The parameter v must not be NULL; it will not be checked.
        """
        pass

    @wrapdll
    def PyLong_FromLong(self, v: c_long) -> py_object:
        """
        Return a new PyLongObject object from v, or NULL on failure.
        """
        pass

    @wrapdll
    def PyLong_AsLong(self, obj: py_object) -> c_long:
        """
        Return a C long representation of obj.
        If obj is not an instance of PyLongObject, first call its __index__ or
        __int__ method (if present) to convert it to a PyLongObject.
        """
        pass

    @wrapdll
    def PyOS_snprintf(self, str: POINTER(c_char), size: c_size_t, format: c_char_p):
        """
        Output not more than size bytes to str according to the format string 
        format and the extra arguments. See the Unix man page snprintf(2).
        """
        pass



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='wrapdll',
    version='0.1.1',
    author='YuvalZ',
    author_email='zanyuval@gmail.com',
    description='Allows using python annotaions to wrap dlls using ctypes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/YuvalZan/wrapdll',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords='ctypes wrapper wrap',
    test_suite="wrapdll.tests",
    python_requires='>=3.6',
)
